/* wptGPGUtil.cpp - GPG helper functions
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <sys/stat.h>

#include "wptErrors.h"
#include "gpgme.h"


#define NROFHEXDIGITS 2  
/* Convert two hexadecimal digits from STR to the value they
   represent.  Returns -1 if one of the characters is not a
   hexadecimal digit.  */
static int
hextobyte (const unsigned char *str)
{
    int val = 0;  
    int i;

    for (i = 0; i < NROFHEXDIGITS; i++) {
	if (*str >= '0' && *str <= '9')
	    val += *str - '0';
	else if (*str >= 'A' && *str <= 'F')
	    val += 10 + *str - 'A';
	else if (*str >= 'a' && *str <= 'f')
	    val += 10 + *str - 'a';
	else
	    return -1;
	if (i < NROFHEXDIGITS - 1)
	    val *= 16;
	str++;
    }
    return val;
}

/* Decode the C formatted string @src and store the result in the
   buffer @destp which is @len bytes long.  If @len is zero, then a
   large enough buffer is allocated with malloc and @destp is set to
   the result.  Currently, @len is only used to specify if allocation
   is desired or not, the caller is expected to make sure that @destp
   is large enough if @len is not zero.  */
gpgme_error_t
gpg_decode_c_string (const char *src, char **destp, size_t len)
{
  char *dest;

  /* Set up the destination buffer.  */
  if (len) {
      if (len < strlen (src) + 1)
	  return gpg_error (GPG_ERR_TOO_SHORT);
      dest = *destp;
  }
  else {      
  /* The converted string will never be larger than the original string.  */
      dest = (char*)malloc (strlen (src) + 1);
      if (!dest)
	  return gpg_error (GPG_ERR_ENOMEM);
      *destp = dest;
  }

  /* Convert the string.  */
  while (*src) {    
      if (*src != '\\') {
	  *(dest++) = *(src++);
	  continue;
      }

      switch (src[1]) {
#define DECODE_ONE(match,result)	\
      case match:			\
	  src += 2;			\
	  *(dest++) = result;		\
	  break;

	  DECODE_ONE ('\'', '\'');
	  DECODE_ONE ('\"', '\"');
	  DECODE_ONE ('\?', '\?');
	  DECODE_ONE ('\\', '\\');
	  DECODE_ONE ('a', '\a');
	  DECODE_ONE ('b', '\b');
	  DECODE_ONE ('f', '\f');
	  DECODE_ONE ('n', '\n');
	  DECODE_ONE ('r', '\r');
	  DECODE_ONE ('t', '\t');
	  DECODE_ONE ('v', '\v');

      case 'x': {
	    int val = hextobyte ((unsigned char*)&src[2]);
	    if (val == -1) { /* Should not happen.  */
		*(dest++) = *(src++);
		*(dest++) = *(src++);
		if (*src)
		  *(dest++) = *(src++);
		if (*src)
		  *(dest++) = *(src++);
	    }
	    else {
		if (!val) {
		    /* A binary zero is not representable in a C string.  */
		    *(dest++) = '\\';
		    *(dest++) = '0'; 
		}
		else 
		    *((unsigned char *) dest++) = val;
		src += 4;
	    }
	}

	default: /* Should not happen.  */
	  {
	    *(dest++) = *(src++);
	    *(dest++) = *(src++);
	  }
        } 
  }
  *(dest++) = 0;
  return 0;
}

/* Replace %foo% entries with its real values.
   Return value: expanded path or NULL on error. */
static char *
expand_path (const char *path)
{
    DWORD len;
    char *p;

    len = ExpandEnvironmentStrings (path, NULL, 0);
    if (!len)
	return NULL;
    len += 1;
    p = (char*)calloc (1, len+1);
    if (!p)
	abort ();
    len = ExpandEnvironmentStrings (path, p, len);
    if (!len) {
	free (p);
	return NULL;
    }
    return p; 
}


/* Read a string from the W32 registry. The directory is given
   in @dir and the name of the value in @name, */
static char *
read_w32_registry (HKEY root_key, const char *dir, const char *name)
{
    HKEY key_handle;
    DWORD n1, nbytes;
    DWORD type;
    char *result = NULL;
    
    if (RegOpenKeyEx (root_key, dir, 0, KEY_READ, &key_handle))
        return NULL; /* no need for a RegClose, so return direct */
    
    nbytes = 1;
    if (RegQueryValueEx (key_handle, name, 0, NULL, NULL, &nbytes))
        goto leave;
    result = (char*)calloc (1, (n1=nbytes+1));
    if (!result)
        abort ();
    if (RegQueryValueEx (key_handle, name, 0, &type, (BYTE*)result, &n1)) {
        free (result); 
	result = NULL;
        goto leave;
    }
    if (type == REG_EXPAND_SZ && strchr (result, '%')) {
	char *p = expand_path (result);
	free (result);
	result = p;
    }
    
leave:
    RegCloseKey (key_handle);
    return result;
}


/* Create a temp file based on the name of @name.
   Return value: handle to the file in case of success. */
static HANDLE
create_tmpfile (const char *name)
{    
    HANDLE out;
    SECURITY_ATTRIBUTES sattr;
    char tmp[300]; 

    memset (&sattr, 0, sizeof sattr);
    sattr.bInheritHandle = TRUE;
    sattr.lpSecurityDescriptor = NULL;
    sattr.nLength = sizeof sattr;

    GetTempPath (sizeof (tmp)-1 - strlen (name)-1, tmp);
    strcat (tmp, name);
    out = CreateFile (tmp, GENERIC_READ|GENERIC_WRITE, 
		      FILE_SHARE_WRITE, &sattr,
		      OPEN_ALWAYS, FILE_FLAG_DELETE_ON_CLOSE, NULL);
    return out;
}


/* Create a pipe with a readable remote end and
   write the data from @dat to the local end.
   Return value: read handle on success. */
static HANDLE
create_in_pipe (const char *dat)
{
    HANDLE r, w;
    SECURITY_ATTRIBUTES sec_attr;
    DWORD n;

    memset (&sec_attr, 0, sizeof sec_attr);
    sec_attr.bInheritHandle = TRUE;
    sec_attr.nLength = sizeof sec_attr;

    if (!CreatePipe (&r, &w, &sec_attr, 4096))
	return NULL;

    WriteFile (w, dat, strlen (dat), &n, NULL);
    CloseHandle (w);

    return r;
}


/* Map the contents of the file handle @out to
   a buffer and return it. */
static char*
map_tmpfile (HANDLE out)
{
    DWORD n;
    char *p;

    FlushFileBuffers (out);
    SetFilePointer (out, 0, NULL, FILE_BEGIN);
    n = GetFileSize (out, NULL);
    p = (char*)calloc (1, n+1);
    if (!p)
	abort ();
    ReadFile (out, p, n, &n, NULL);
    p[n] = 0;
    return p;
}


/* Create a process from the command line in @cmd.
   If @out is != NULL, the output of the process will
   be redirected to @out. If @in is != NULL the input
   will be read from @in.
   Return value: 0 on success. */
static int
create_process (const char *cmd, HANDLE in, HANDLE out)
{
    STARTUPINFO si;
    PROCESS_INFORMATION pi;

    memset (&si, 0, sizeof (si));
    si.cb = sizeof si;
    if (in || out)
	si.dwFlags = STARTF_USESTDHANDLES|STARTF_USESHOWWINDOW;
    if (out)
	si.hStdOutput = out;
    if (in)
	si.hStdInput = in;
    si.wShowWindow = SW_HIDE;
    if (!CreateProcess (NULL, (char*)cmd, NULL, NULL, TRUE, 0, 
		        NULL, NULL, &si, &pi)) {
	log_debug ("create_process: CreateProcess() failed ec=%d\r\n",
		   (int)GetLastError ());
	return -1;
    }
    WaitForSingleObject (pi.hProcess, INFINITE);
    CloseHandle (pi.hProcess);
    return 0;
}


/* Export a GPG secret key given by @keyid into the file @outfile.
   Return value: 0 on success. */
gpgme_error_t
gpg_export_seckey (const char *keyid, const char *outfile)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    struct stat st;
    char *p;
    char *cmd;

    p = read_w32_registry (HKEY_CURRENT_USER, 
			   "Software\\GNU\\GnuPG", "gpgProgram");
    if (!p)
	return gpg_error (GPG_ERR_INV_ARG);
    cmd = (char*)calloc (1, strlen (p) + strlen (keyid) 
			    + strlen (outfile) + 64 + 2);
    if (!cmd)
	abort ();
    sprintf (cmd, "%s --yes --output \"%s\" --export-secret-key %s", 
	     p, outfile, keyid);
    if (create_process (cmd, NULL, NULL))
	err = gpg_error (GPG_ERR_INTERNAL);

    if (stat (outfile, &st) == -1 || st.st_size == 0)
	err = gpg_error (GPG_ERR_NO_DATA);

    free (p);
    free (cmd);
    return err;
}


/* If EXPORTFLAG is 1, export the ownertrust data to the
   buffer DATA. Otherwise import the ownertrust data from DATA.
   Return value: 0 on success. */
gpgme_error_t
gpg_manage_ownertrust (char **data, int exportflag)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    HANDLE out = NULL, in = NULL;
    char *p;
    char *cmd;

    p = read_w32_registry (HKEY_CURRENT_USER, 
			   "Software\\GNU\\GnuPG", "gpgProgram");
    if (!p)
	return gpg_error (GPG_ERR_INV_ARG);

    cmd = (char*)calloc (1, strlen (p) + 64 + 1);
    if (!cmd)
	abort ();  
    sprintf (cmd, "%s %s", p, 
	     exportflag? "--export-ownertrust" : "--import-ownertrust");

    if (exportflag)
	out = create_tmpfile ("gpg_ot_out");
    else {
	DWORD nw;
	in = create_tmpfile ("gpg_ot_in");
	WriteFile (in, *data, strlen (*data), &nw, NULL);
	FlushFileBuffers (in);
	/* XXX: need a rewind? */
    }
    if (create_process (cmd, in, out))
	err = gpg_error (GPG_ERR_INTERNAL);

    free (p);
    free (cmd);

    if (in)
	CloseHandle (in);
    if (out) {
	*data = map_tmpfile (out);
	CloseHandle (out);
    }
    return err;
}


/* Call gpg --rebuild-keydb-caches to speed up signature listings. */
gpgme_error_t
gpg_rebuild_cache (char **r_inf)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    HANDLE out = NULL;
    char *p;
    char *cmd;

    p = read_w32_registry (HKEY_CURRENT_USER, 
			   "Software\\GNU\\GnuPG", "gpgProgram");
    if (!p)
	return gpg_error (GPG_ERR_INV_ARG);
    cmd = (char*)calloc (1, strlen (p) + 64);
    if (!cmd)
	abort ();
    sprintf (cmd, "%s --logger-fd=1 --rebuild-keydb-caches", p);

    if (r_inf)
	out = create_tmpfile ("gpg_rebuild_cache");

    if (create_process (cmd, NULL, out))
	err = gpg_error (GPG_ERR_INTERNAL);

    if (r_inf)
	*r_inf = map_tmpfile (out);
    if (out)
	CloseHandle (out);
    free (p);
    free (cmd);
    return 0;
}


/* Call gpg --version to retrieve the 'about' information. */
gpgme_error_t
gpg_get_version (char **r_inf)
{
    gpgme_error_t err= gpg_error (GPG_ERR_NO_ERROR);
    HANDLE out;    
    char *p, *cmd;

    p =read_w32_registry (HKEY_CURRENT_USER, 
			  "Software\\GNU\\GnuPG", "gpgProgram");
    if (!p)
	return gpg_error (GPG_ERR_INV_ARG);
    cmd = (char*)calloc (1, strlen (p) + 32);
    if (!cmd)
	abort ();
    sprintf (cmd, "%s --version", p);

    out = create_tmpfile ("gpg_out");
    if (create_process (cmd, NULL, out))
	err = gpg_error (GPG_ERR_INTERNAL);

    free (p);
    free (cmd);

    *r_inf = map_tmpfile (out);
    CloseHandle (out);    
    return err;
}


/* Return the colon file output of the given file @fname in @r_out.
   Return value: 0 on success. */ 
gpgme_error_t
gpg_import_key_list (const char *fname, char **r_out)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    char *cmd, *p;
    HANDLE out;

    p = read_w32_registry (HKEY_CURRENT_USER, 
			   "Software\\GNU\\GnuPG", "gpgProgram");
    if (!p)
	return gpg_error (GPG_ERR_INV_ARG);

    cmd = (char*)calloc (1, strlen (p) + strlen (fname) + 2+2 + 64);
    if (!cmd)
	abort ();
    sprintf (cmd, "%s --fixed-list-mode --with-colons \"%s\"", p, fname);

    out = create_tmpfile ("gpg_keys");
    if (create_process (cmd, NULL, out))
	err = gpg_error (GPG_ERR_INTERNAL);

    free (p);
    free (cmd);

    *r_out = map_tmpfile (out);
    CloseHandle (out);
    return err;
}

char* 
generate_revoc_input (int code, const char *cmt, const char *pass)
{
    const char *fmt;
    char *p;
    size_t n;

    fmt = "Y\n"	    /*  gen_revoke.okay */
	  "%d\n"    /* ask_revocation_reason.code */
	  "%s\n"    /* ask_revocation_reason.text */
	  "%s"	    /* text != NULL '\n' otherwise '' */
	  "Y\n"	    /* ask_revocation_reason.okay */
	  "%s\n";   /* passphrase.enter. */
    n = strlen (fmt) + 32;
    if (pass)
	n += strlen (pass) + 1;
    if (cmt)
	n += strlen (cmt) + 1;
    p = (char*)calloc (1, n+1);
    if (!p)
	abort ();
    sprintf (p, fmt, code, cmt? cmt : "", cmt? "\n" : "", pass? pass : "");
    return p;
}
    

/* Generate a revocation certificate for the key with the keyid @keyid.
   @inp_data contains all needed data to answer the questions of the
   command handler. Each separate with a '\n'.
   @r_revcert contains the revocation cert on success.
   Return value: 0 on success. */
gpgme_error_t
gpg_revoke_key (const char *inp_data, const char *keyid, char **r_revcert)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    char *rcrt;
    char *cmd, *p;
    HANDLE in, out;
    
    p = read_w32_registry (HKEY_CURRENT_USER, 
			   "Software\\GNU\\GnuPG", "gpgProgram");
    if (!p)
	return gpg_error (GPG_ERR_INV_ARG);

    cmd = (char*)calloc (1, strlen (p) + strlen (keyid)+1 + 128);
    if (!cmd)
	abort ();
    sprintf (cmd, "%s --pgp7 --command-fd=0 --status-fd=2 --gen-revoke %s", 
	     p, keyid);

    in = create_in_pipe (inp_data);
    out = create_tmpfile ("gpg_revcert");
    if (create_process (cmd, in, out)) {
	*r_revcert = NULL;
	err = gpg_error (GPG_ERR_INTERNAL);
    }
    else {
	rcrt = map_tmpfile (out);
	*r_revcert = rcrt;
    }

    free (p);
    free (cmd);

    CloseHandle (in);
    CloseHandle (out);
    return err;
}


/* Return the validity of the user attribute, informerly known 
   as photo-ID. If no uat was found, return 0 for unknown. */
gpgme_error_t
get_uat_validity (const char *keyid, gpgme_validity_t *r_valid)
{
    gpgme_error_t err = gpg_error (GPG_ERR_NO_ERROR);
    HANDLE out;
    char *p, *cmd;
    char *uat;

    *r_valid = GPGME_VALIDITY_UNKNOWN;
    p = read_w32_registry (HKEY_CURRENT_USER, 
			   "Software\\GNU\\GnuPG", "gpgProgram");
    if (!p)
	return gpg_error (GPG_ERR_INV_ARG);

    cmd = (char*)calloc (1, strlen (p) + strlen (keyid)+1 + 128);
    if (!cmd)
	abort ();
    sprintf (cmd, "%s --with-colons --fixed-list-mode --list-keys \"%s\"",
	     p, keyid);

    out = create_tmpfile ("gpg_keys");
    if (create_process (cmd, NULL, out))
	err = gpg_error (GPG_ERR_INTERNAL);

    free (p);
    free (cmd);
    
    p = map_tmpfile (out);
    if ((uat = strstr (p, "uat:"))) {
	switch (*(uat+4)) {
	case 'm': *r_valid = GPGME_VALIDITY_MARGINAL; break;
	case 'f':
	case 'u': *r_valid = GPGME_VALIDITY_FULL; break;
	default: *r_valid = GPGME_VALIDITY_UNDEFINED; break;
	}
    }

    free (p);    
    CloseHandle (out);
    return err;
}


#if 0
/* Extract all recipients from the file @file.
   Return value: 0 on success. */
static int
do_get_recipients (const char *file, gpgme_recipient_t *r_list)
{
    gpgme_recipient_t l;
    PACKET *pkt;
    gpg_iobuf_t inp = NULL;
    armor_filter_context_t afx;
    int rc = 0, quit=0;

    if (!file || !r_list) {
	log_debug ("do_list_packets: !r_list || !file");
	return -1;
    }

    *r_list=NULL;
    inp = gpg_iobuf_open (file);
    if (!inp)
	return WPTERR_FILE_OPEN;
    gpg_iobuf_ioctl (inp, 3, 1, NULL); /* disable cache */
    if (gpg_use_armor_filter (inp)) {
	memset (&afx, 0, sizeof (afx));
	gpg_iobuf_push_filter (inp, gpg_armor_filter, &afx);
    }
    pkt = (PACKET *)calloc(1, sizeof *pkt);
    gpg_init_packet (pkt);
    while (!quit && (rc = gpg_parse_packet (inp, pkt)) != -1) {
	switch (pkt->pkttype) {
	case PKT_PUBKEY_ENC:
	    {PKT_pubkey_enc *enc = pkt->pkt.pubkey_enc;
	    if (!enc)
		break;
	    l = calloc (1, sizeof *l);
	    l->keyid = calloc (1, 16+1);
	    _snprintf (l->keyid, 16, "%08lX%08lX", enc->keyid[0], enc->keyid[1]);
	    l->pubkey_algo = enc->pubkey_algo;
	    l->status = 0;
	    l->next = (*r_list);
	    *r_list = l;
	    break;}

	case PKT_ENCRYPTED:	
	case PKT_ENCRYPTED_MDC:
	case PKT_COMPRESSED:
	case PKT_PUBLIC_KEY:
	case PKT_SECRET_KEY:
	    quit = 1;
	    break;
	}
	gpg_free_packet (pkt);
	gpg_init_packet (pkt);
    }
    gpg_iobuf_close (inp);
    safe_free (pkt);
    return 0;
}
#endif
