/* wptKeyRevokersDlg.cpp - Designated Revoker Keys
 *	Copyright (C) 2002, 2003, 2005 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>

#include "resource.h"
#include "wptCommonCtl.h"
#include "wptNLS.h"
#include "wptTypes.h"
#include "wptGPG.h"
#include "wptCommonCtl.h"
#include "wptContext.h"
#include "wptDlgs.h"
#include "wptErrors.h"
#include "wptKeyserver.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptKeyEdit.h"
#include "wptKeylist.h"


/* Initialize the listview given in @ctrl and return it in @r_lv. */   
static int
revokelist_build (listview_ctrl_t *r_lv, HWND ctrl)
{
    listview_ctrl_t lv;
    struct listview_column_s rlist[] = {
	{ 0,  55, (char *)_("Algorithm") },
	{ 1,  80, (char *)_("Key ID") },
	{ 2, 160, (char *)_("Name") },
	{0}
    };
    int i, rc;

    rc = listview_new (&lv);
    if (rc)
	return rc;
    lv->ctrl = ctrl;
    lv->items = DIM( rlist )-1;
    for (i = 0; i < lv->items; i++)
	listview_add_column (lv, &rlist[i]);
    listview_set_ext_style (lv);
    *r_lv = lv;
    return 0;
}


/* Initialize the listview @ctrl with all designated revokers of the
   key @key. @ctx is used to retrieve the rev key from the cache to
   show the user-id.
   Return value: the initialized listview or NULL on error. */
static listview_ctrl_t
revokelist_init (HWND ctrl, gpgme_key_t key, gpg_keycache_t ctx)
{    
    gpgme_error_t err;
    gpgme_key_t revkey;    
    GpgKeyEdit *ke;    
    gpg_desig_rev_t rev=NULL, r;
    listview_ctrl_t lv;    
    const char *id;

    ke = new GpgKeyEdit (key);
    if (!ke)
	BUG (NULL);
    err = ke->getDesignatedRevoker (&rev);
    if (err) {
	delete ke;
	msg_box (NULL, gpgme_strerror (err), _("Designated Key Revokers"), MB_ERR);
	return NULL;
    }

    if (revokelist_build (&lv, ctrl))
	BUG (0);
    for (r = rev; r; r = r->next) {
	if (get_pubkey (r->fpr+32, &revkey))
	    id = _("user ID not found");
	else
	    id = revkey->uids->name;
	listview_add_item (lv, "");
	listview_add_sub_item (lv, 0, 0, get_key_pubalgo (r->pubkey_algo));
	listview_add_sub_item (lv, 0, 1, r->fpr+32);
	listview_add_sub_item (lv, 0, 2, id);
    }
        
    delete ke;
    gpg_desig_rev_release (rev);
    return lv;
}


/* Dialog box procedure to show designated revokers. */
BOOL CALLBACK
key_revokers_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv;
    gpgme_key_t key;
    char buf[128], keyid[32];
    int rc;
    
    switch( msg ) {
    case WM_INITDIALOG:
	gpg_keycache_t ctx;
	key = (gpgme_key_t)lparam;
	if (!key)
	    dlg_fatal_error( dlg, "Could not get dilaog param" );
	ctx = keycache_get_ctx (1);
	if (!ctx)
	    BUG (dlg);
	lv = revokelist_init (GetDlgItem (dlg, IDC_KEYREVOKERS_LIST), key, ctx);

	SetDlgItemText (dlg, IDC_KEYREVOKERS_INF, _("Designated Revoker Keys"));
	SetWindowText (dlg, _("Key Revokers"));
	SetForegroundWindow (dlg);
	return TRUE;

    case WM_NOTIFY:
	NMHDR *notify;
	notify = (NMHDR *)lparam;
	if (notify && notify->code == NM_DBLCLK && 
	    notify->idFrom == IDC_KEYREVOKERS_LIST) {
	    int idx = listview_get_curr_pos (lv);
	    listview_get_item_text (lv, idx, 1, keyid, DIM (keyid)-1);
	    listview_get_item_text (lv, idx, 2, buf, DIM (buf)-1);
	    if (!strcmp( buf, _("user ID not found"))) {
		int id = log_box (_("Designated Key Revokers"), MB_YESNO|MB_INFO,
				  _("Do you want to retrieve 0x%s via the default keyserver?"), keyid);
		if (id == IDNO)
		    break;
		rc = hkp_recv_key (dlg, default_keyserver, 
				   default_keyserver_port, buf, 0, 0);
		if( !rc ) {
		    keycache_reload (dlg);
		    if (get_pubkey (keyid, &key))
			BUG (NULL);
		    if (key->uids && key->uids->name)
			listview_add_sub_item (lv, idx, 2, key->uids->name);
		}
	    }
	    else {
		struct winpt_key_s k;
		memset (&k, 0, sizeof k);
		k.keyid = keyid;
		DialogBoxParam (glob_hinst, (LPCTSTR)IDD_WINPT_KEYPROPS, dlg,
				keyprops_dlg_proc, (LPARAM)&k);
	    }
	}
	break;

    case WM_DESTROY:
	if (lv) {
	    listview_release (lv);
	    lv = NULL;
	}
	return FALSE;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    EndDialog (dlg, TRUE);
	    break;
	}
	break;
    }

    return FALSE;
}
