#include <windows.h>
#include <malloc.h>
#include <string.h>
#include <stdio.h>

#include "w32gpgme.h"


/* Implement a word based line break. @inp is the input buffer
   and @wraplen is the maximal line length.
   Return value: the wrapped buffer on success. */
char*
wrap_lines (char *inp, size_t inlen, size_t wraplen, size_t *r_len)
{
    char *out;
    char *p, *inp_ptr;
    int end_is_le=0, add_le=0;
    size_t n;

    inp[inlen] = 0;
    if (inp[inlen-2] == '\r' && inp[inlen-1] == '\n')
	end_is_le = 1;
    out = calloc (1, 4*inlen/3+2+1);
    if (!out)
	return NULL;
    n = 0;
    inp_ptr = inp;
    while ((p = strsep (&inp_ptr, " \n"))) {
	if (strlen (p) == 0) {
	    strcat (out, " ");
	    n++;
	    continue;
	}
	/* if we find an existing \r\n pair, we generate a break
	   at the given position and reset the counter. */
	if (p[strlen (p)-1] == '\r') {
	    p[strlen (p)-1]=0;
	    add_le=1;
	}
	else
	    add_le=0;
	if (n + strlen (p) > wraplen) {
	    strcat (out, "\r\n");
	    n = 0;
	}
	else if (n > 0 || add_le) {
	    strcat (out, " ");
	    n++;
	}

	n += strlen (p);
	strcat (out, p);
	if (add_le) {
	    strcat (out, "\r\n");
	    n = 0;
	}
    }
    if (end_is_le)
	strcat (out, "\r\n");
    *r_len = strlen (out);
    return out;
}


/* Wrap the lines of @r_dh with a line length of @wraplen.
   @r_dh will be released and on success it contains the
   handle to the wrapped data.
   Return value: 0 on success. */
gpgme_error_t
gpg_data_wrap_lines (gpgme_data_t *r_dh, size_t wraplen)
{	
    gpgme_error_t err;
    gpgme_data_t mdh;
    char *raw = NULL;
    char *p = NULL;
    size_t nlen;
    
    err = gpgme_data_new (&mdh);
    if (err)
        return err;

    raw = gpgme_data_release_and_get_mem (*r_dh, &nlen);
    if (!raw) {
	gpgme_data_release (mdh);
	return gpg_error (GPG_ERR_INV_ARG);
    }
	
    p = wrap_lines (raw, nlen, wraplen, &nlen);
    if (p) {    
	gpgme_data_write (mdh, p, nlen);    
	gpgme_data_rewind (mdh);    
	free (p);
    }
    gpgme_free (raw);
    *r_dh = mdh;
    return 0;
}


/* Prepend '> ' to each line into @r_dh. On success @r_dh
   contains a handle to the quoted data. 
   Return value: 0 on success. */
gpgme_error_t
gpg_data_mail_quote (gpgme_data_t *r_dh)
{
    gpgme_data_t dh;
    char buf[128];

    if (!*r_dh)
	return gpg_error (GPG_ERR_INV_ARG);
    gpgme_data_new (&dh);
    while (gpg_data_readline (*r_dh, buf, 127)) {
	gpgme_data_write (dh, "> ", 2);
	gpgme_data_write (dh, buf, strlen (buf));
    }
    gpgme_data_release (*r_dh);
    *r_dh = dh;
    return 0;
}


/* Extract the plaintext data from the escaped data object @sig.
   The plaintxt is stored in @r_plain.
   Return value: 0 on success. */
gpgme_error_t
gpg_data_extract_plaintext (gpgme_data_t sig, gpgme_data_t *r_plain)
{
    gpgme_data_t plain;
    gpgme_error_t err;
    char line[128+32];
    int pos = 0;
    int sig_begin = 0;
    
    if (r_plain)
	*r_plain = NULL;
    err = gpgme_data_new (&plain);
    if (err)
        return err;
    
    while (gpg_data_readline (sig, line, 128) > 0) {
        if (!strncmp (line, "-----BEGIN PGP SIGNED MESSAGE", 29) ||
            !strncmp (line, "Version:", 8) ||
            !strncmp (line, "Comment:", 8) ||
            !strncmp (line, "Charset:", 8) ||
            !strncmp (line, "Hash:", 5) ||
	    !strncmp (line, "MessageID", 9))
            continue;
        if (strlen (line) <= 2)
            break; /* parsed all headers, now we reached the body */
    }
    /* XXX handle multi dash escaped sequences */
    while (gpg_data_readline (sig, line, 128) > 0 ) {
        if (!strncmp( line, "-----BEGIN PGP SIGNATURE", 24))
            break; /* end of plaintext */
        if (!strncmp (line, "- -", 3))
            pos = 2;
	gpgme_data_write (plain, line+pos, strlen (line+pos));
        pos = 0;
    }
    gpgme_data_write (plain, "", 1);
    if (r_plain)
	*r_plain = plain;
    else
	gpgme_data_release (plain);
    return err;
}


/* Release the data in @dh and store the contents in the file @fname.
   Return value: 0 on success. */
gpgme_error_t
gpg_data_release_and_set_file (gpgme_data_t dh, const char *fname)
{
    char *p = NULL;
    FILE *fp;
    size_t n;
    
    fp = fopen (fname, "wb");
    if (fp == NULL)
	return gpg_error (GPG_ERR_ENOENT);
    
    p = gpgme_data_release_and_get_mem (dh, &n);
    if (p) {
        fwrite (p, 1, n, fp);
        fflush (fp);
	memset (p, 0xFF, n);
	gpgme_free (p);
    }
    fclose (fp);
    return 0;
}


/* Try to read a line, terminated by \n from the given
   gpgme data object @dh into @line. 
   Return value: numbers of chars read. */
size_t
gpg_data_readline (gpgme_data_t dh, char *line, size_t nbytes)
{
    char ch = 0;
    int nread = 0, pos = 0;

    if (!dh)
	return 0;
    
    memset (line, 0, nbytes);
    while ((nread=gpgme_data_read (dh, &ch, 1)) > 0) {
        if (!nread)
            break;
        if (ch == '\n') {
            line[pos++] = ch;
            line[pos++] = '\0';
            break;
        }
        line[pos++] = ch;
        if (pos > (int)nbytes) {
            line[pos++] = '\0';
            break;
        }
    }
    
    return pos;
}
